/*
 * This is the MIT license, see also http://www.opensource.org/licenses/mit-license.html
 *
 * Copyright (c) 2001 Brian Pitcher
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

// $Header: /Users/cpa/cvsroot/weblech-0.0.3-AJ/src/weblech/spider/Spider.java,v 1.2 2010/08/20 12:48:55 cpa Exp $

package weblech.spider;

import weblech.util.Logger;
import weblech.util.Log4j;

import atomicsets.OrderedLock;
import java.util.*;
import java.io.*;
import java.net.URL;

import org.apache.log4j.Category;

public class Spider extends Logger implements Constants, atomicsets.Atomic
{
	/*atomicset(S)*/
	
    public final OrderedLock getLockForS() {
		return $lock_S;
	}

	public final OrderedLock getLock() {
		return this.getLockForS();
	}

	/** Config for the spider */
    private SpiderConfig config;
    /**
     * Download queue.
     * Thread safety: To access the queue, first synchronize on it.
     */
    /*atomic(S)*/ private DownloadQueue queue;
    /**
     * Set of URLs currently being downloaded by Spider threads.
     * Thread safety: To access the set, first synchronize on it.
     */
    /*atomic(S)*/ private Set urlsDownloading;
    /**
     * Number of downloads currently taking place.
     * Thread safety: To modify this value, first synchronize on
     *                the download queue.
     */
    /*atomic(S)*/ private int downloadsInProgress;

    /** Time we last checkpointed. */
    private long lastCheckpoint;

    public Spider(SpiderConfig config) {
		this(config, new OrderedLock());
	}

	public Spider(SpiderConfig config, OrderedLock S)
    {
        $lock_S = S;
		this.config = config;
        queue = new DownloadQueue(config);
        queue.queueURL(new URLToDownload(config.getStartLocation(), 0));
        // TODD: Moved to the UDorSWrapper class
//        urlsDownloadedOrScheduled = new HashSet();
        urlsDownloading = new HashSet();
        downloadsInProgress = 0;
        lastCheckpoint = 0;
    }
    
    //TODD: Added this method 
    public boolean downloadsInProgress_internal() { return downloadsInProgress > 0; }

	public boolean downloadsInProgress() {
		synchronized (this.$lock_S) {
			return downloadsInProgress > 0;
		}
	}

    //TODD: Making this method public so it can be called from SpiderRunnable
    public void checkpointIfNeeded_internal()
    {
        if(config.getCheckpointInterval() == 0)
        {
            return;
        }

        if(System.currentTimeMillis() - lastCheckpoint > config.getCheckpointInterval())
        {
            checkpointIfNeededAux_internal();
        }
    }

	public void checkpointIfNeeded() {
		synchronized (this.$lock_S) {
			if(config.getCheckpointInterval() == 0)
			{
			    return;
			}
			if(System.currentTimeMillis() - lastCheckpoint > config.getCheckpointInterval())
			{
			    checkpointIfNeededAux_internal();
			}
		}
	}

	public void checkpointIfNeededAux_internal() {
//		synchronized(queue)
		{
		    if(System.currentTimeMillis() - lastCheckpoint > config.getCheckpointInterval())
		    {
		        writeCheckpoint_internal();
		        lastCheckpoint = System.currentTimeMillis();
		    }
		}
	}

	public void checkpointIfNeededAux() {
		{
			synchronized (this.$lock_S) {
				if(System.currentTimeMillis() - lastCheckpoint > config.getCheckpointInterval())
				{
				    writeCheckpoint_internal();
				    lastCheckpoint = System.currentTimeMillis();
				}
			}
		}
	}
    
 

    private void writeCheckpoint_internal()
    {
        _logClass.debug("writeCheckpoint()");
        try
        {
            FileOutputStream fos = new FileOutputStream("spider.checkpoint", false);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            oos.writeObject(queue);
            oos.writeObject(urlsDownloading);
            oos.close();
        }
        catch(IOException ioe)
        {
            _logClass.warn("IO Exception attempting checkpoint: " + ioe.getMessage(), ioe);
        }
    }

	private void writeCheckpoint() {
		synchronized (this.$lock_S) {
			_logClass.debug("writeCheckpoint()");
			try
			{
			    FileOutputStream fos = new FileOutputStream("spider.checkpoint", false);
			    ObjectOutputStream oos = new ObjectOutputStream(fos);
			    oos.writeObject(queue);
			    oos.writeObject(urlsDownloading);
			    oos.close();
			}
			catch(IOException ioe)
			{
			    _logClass.warn("IO Exception attempting checkpoint: " + ioe.getMessage(), ioe);
			}
		}
	}

    public void readCheckpoint_internal()
    {
        try
        {
            FileInputStream fis = new FileInputStream("spider.checkpoint");
            ObjectInputStream ois = new ObjectInputStream(fis);
            queue = (DownloadQueue) ois.readObject();
            urlsDownloading = (Set) ois.readObject();
            queue.queueURLs(urlsDownloading);
            urlsDownloading.clear();
        }
        catch(Exception e)
        {
            _logClass.error("Caught exception reading checkpoint: " + e.getMessage(), e);
        }
    }

	public void readCheckpoint() {
		synchronized (this.$lock_S) {
			try
			{
			    FileInputStream fis = new FileInputStream("spider.checkpoint");
			    ObjectInputStream ois = new ObjectInputStream(fis);
			    queue = (DownloadQueue) ois.readObject();
			    urlsDownloading = (Set) ois.readObject();
			    queue.queueURLs(urlsDownloading);
			    urlsDownloading.clear();
			}
			catch(Exception e)
			{
			    _logClass.error("Caught exception reading checkpoint: " + e.getMessage(), e);
			}
		}
	}


	public void queueURLs_internal(ArrayList u2dsToQueue) {
		// synchronized(queue)
		{
		    queue.queueURLs(u2dsToQueue);
		    downloadsInProgress--;
		}
	}

	public void queueURLs(ArrayList u2dsToQueue) {
		{
			synchronized (this.$lock_S) {
				queue.queueURLs(u2dsToQueue);
				downloadsInProgress--;
			}
		}
	}

	//TODD: Extracted from run()
	public URLToDownload getNextURL_internal() {
		URLToDownload nextURL;
		// synchronized(queue)
		{
		    nextURL = queue.getNextInQueue();
		    downloadsInProgress++;
		}
		// synchronized(urlsDownloading)
		{
		    urlsDownloading.add(nextURL);
		}
		return nextURL;
	}

	public URLToDownload getNextURL() {
		URLToDownload nextURL;
		synchronized (this.$lock_S) {
			// synchronized(queue)
			{
			    nextURL = queue.getNextInQueue();
			    downloadsInProgress++;
			}
			// synchronized(urlsDownloading)
			{
			    urlsDownloading.add(nextURL);
			}
		}
		return nextURL;
	}



    /**
     * Get the size of the download queue in a thread-safe manner.
     * TODD: Changed from private to public
     */
    public int queueSize_internal()
    {
       // synchronized(queue)
        {
            return queue.size();
        }
    }

	/**
	 * Get the size of the download queue in a thread-safe manner. TODD: Changed from private to public
	 */
	public int queueSize() {
		{
			synchronized (this.$lock_S) {
				return queue.size();
			}
		}
	}

    /**
     * Get a URL, and return new URLs that are referenced from it.
     *
     * @return A List of URL objects.
     */
    // TODD: Made the method public
    public List downloadURL_internal(URLToDownload url, URLGetter urlGetter, HTMLParser htmlParser)
    {
        _logClass.debug("downloadURL(" + url + ")");

        // TODD: moved this block from the run() method to here
		// synchronized(urlsDownloading)
		{
		    urlsDownloading.remove(url);
		}
        
        // Bail out early if image and already on disk
        URLObject obj = new URLObject(url.getURL(), config);
        if(obj.existsOnDisk())
        {
            if(config.refreshHTMLs() && (obj.isHTML() || obj.isXML()))
            {
                _logClass.info("Q: [" + queue + "] " + url);
                obj = urlGetter.getURL(url);
            }
            else if(config.refreshImages() && obj.isImage())
            {
                _logClass.info("Q: [" + queue + "] " + url);
                obj = urlGetter.getURL(url);
            }
        }
        else
        {
            _logClass.info("Q: [" + queue + "] " + url);
            obj = urlGetter.getURL(url);
        }

        if(obj == null)
        {
            return new ArrayList();
        }

        if(!obj.existsOnDisk())
        {
            obj.writeToFile();
        }

        if(obj.isHTML() || obj.isXML())
        {
            return htmlParser.parseLinksInDocument(url.getURL(), obj.getStringContent());
        }
        else if(obj.isImage())
        {
            return new ArrayList();
        }
        else
        {
            _logClass.warn("Unsupported content type received: " + obj.getContentType());
            _logClass.info("URL was " + url);
            return new ArrayList();
        }
    }

	/**
	 * Get a URL, and return new URLs that are referenced from it.
	 * @return  A List of URL objects.
	 */
	public List downloadURL(URLToDownload url, URLGetter urlGetter,
			HTMLParser htmlParser) {
		synchronized (this.$lock_S) {
			_logClass.debug("downloadURL(" + url + ")");
			// TODD: moved this block from the run() method to here
			// synchronized(urlsDownloading)
			{
			    urlsDownloading.remove(url);
			}
			// Bail out early if image and already on disk
			URLObject obj = new URLObject(url.getURL(), config);
			if(obj.existsOnDisk())
			{
			    if(config.refreshHTMLs() && (obj.isHTML() || obj.isXML()))
			    {
			        _logClass.info("Q: [" + queue + "] " + url);
			        obj = urlGetter.getURL(url);
			    }
			    else if(config.refreshImages() && obj.isImage())
			    {
			        _logClass.info("Q: [" + queue + "] " + url);
			        obj = urlGetter.getURL(url);
			    }
			}
			else
			{
			    _logClass.info("Q: [" + queue + "] " + url);
			    obj = urlGetter.getURL(url);
			}
			if(obj == null)
			{
			    return new ArrayList();
			}
			if(!obj.existsOnDisk())
			{
			    obj.writeToFile();
			}
			if(obj.isHTML() || obj.isXML())
			{
			    return htmlParser.parseLinksInDocument(url.getURL(), obj.getStringContent());
			}
			else if(obj.isImage())
			{
			    return new ArrayList();
			}
			else
			{
			    _logClass.warn("Unsupported content type received: " + obj.getContentType());
			    _logClass.info("URL was " + url);
			    return new ArrayList();
			}
		}
	}

    
    //TODD: Extracted from start()
    public void setup_internal() {
      _logClass.info("Starting Spider thread");
    }

	public void setup() {
		synchronized (this.$lock_S) {
			_logClass.info("Starting Spider thread");
		}
	}
    
    //TODD: Extracted from run()
    public void cleanup_internal() {
        _logClass.info("Spider thread stopping");
    }

	public void cleanup() {
		synchronized (this.$lock_S) {
			_logClass.info("Spider thread stopping");
		}
	}

	protected final OrderedLock $lock_S;

}
